"""@package XTRESTAPI
    A Python3 class that allows access to all of the functionality in the
    Xpert-Timer XTWEB/XTCLOURSERVER API.

    This package attempts to stay on top of changes to the API and allow an
    easy to user interface with the XTWEB API. The API does no formatting
    for the user. The API responses will be returned as the raw JSON.

    This API was built with the developer in mind and should allow a developer
    to build applications around the XTWEB API without having to deal with
    accessing and managing the requests and responses.
"""

import requests             # Needs to be installed by using:
                            # > python -m pip install requests

import datetime
import json
import uuid

from .xtconst import *
from .URLs import URLs

XTRESTAPILEVEL = 1   # Version of the XTREST-API

class XTRESTAPI:
    """The XTRESTAPI class providing blackbox use of the Xpert-Timer / XTWEB API.

    This is the main class of the API module. This should be the only class used in
    applictions built around the API. The XTRESTAPI class allows access to the GET and
    POST requests supported by Xpert-Timer / XTWEB.

    """
    def __init__(self, api_url, api_key, api_username, api_password):
        """XTRESTAPI constructor. Sets the response format on all of the URLs and
            the oauth/client keys required to access the API.

            Parameters
                @param self - the object pointer
                @param api_key - GUID of API-Key on XTCloudserver
        """
        self.url = URLs(api_url)

        self.api_url = api_url
        self.api_key = api_key

        self.api_username = api_username             # Username for BasicAuth on Server
        self.api_password = api_password             # Password for BasicAuth on Server

        self.HEADERS = {}

        self.auth_time = None
        self.auth_token = None

        self.last_error = None
        self.last_error_message = None
        

# ==========================================================================================================================

    def __get_data(self, url, params = None):
        """A private method to return the requested data in the requested format
            for a given URL.
            @param self - the object pointer
            @param url - API URL to access
            @param params - params for the HTTP request
        """
        return self.__check_error(requests.get(url, json=params, headers=self.HEADERS, auth=(self.api_username,self.api_password)))


    def __submit_post(self, url, data):
        """A private method to submit a post request to the XTCloud server
            @param self - the object pointer
            @param url - API URL to access
            @param data - payload for the HTTP request
        """
        return self.__check_error(requests.post(url, json=data, headers=self.HEADERS, auth=(self.api_username,self.api_password)))

    def __put_data(self, url, data):
        """A private method to submit a post request to the XTCloud server
            @param self - the object pointer
            @param url - API URL to access
            @param data - payload for the HTTP request
        """
        return self.__check_error(requests.put(url, json=data, headers=self.HEADERS, auth=(self.api_username,self.api_password)))

    def __delete(self, url):
        """A private method to submit a post request to the XTCloud server
            @param self - the object pointer
            @param url - API URL to access
        """
        return self.__check_error(requests.delete(url, headers=self.HEADERS, auth=(self.api_username,self.api_password)))

    def __call_jrpc(self, methodname, params, id = 1):
        """A private method to call a jrpc method on the XTCloud server
            @param self - the object pointer
            @param url - API URL to access
            @param params - array of params for the JRPC-Call request
        """
        PARAMS = {
        "jsonrpc":"2.0",
        "id":id,
        "method":methodname,
        "params":params
        }
        # print(f"__call_jrpc()", PARAMS) # Debug
        return self.__check_error_jrpc(requests.post(self.url.jrpc_url(), json=PARAMS, headers=self.HEADERS, auth=(self.username,self.password)))


    def __check_error(self, response):
        """Check response
            @param self - the object pointer
            @param response - response from the XTCloudserver
        """
        result = (response.status_code >= 200) and (response.status_code <= 229)
        if not result:
            self.last_error = response.status_code
            self.last_error_message = response.reason
        return response

    def __check_error_jrpc(self, response):
        """Check response
            @param self - the object pointer
            @param response - response from the XTCloudserver
        """
        result = (response.status_code >= 200) and (response.status_code <= 229)
        if not result:
            self.last_error = response.status_code
            self.last_error_message = response.reason
        return response

# ==========================================================================================================================

    def CreateGUID(self):
        """Creates a GUID as a STRING
            @param self - the object pointer
        """
        return "{" + str(uuid.uuid4()) + "}"

# ==========================================================================================================================

    def check(self, response, showstatus = False):
        """Check response
            @param self - the object pointer
            @param response - response from the XTCloudserver
        """
        result = (response.status_code >= 200) and (response.status_code <= 229)
        if not result:
            self.last_error = response.status_code
            self.last_error_message = response.reason
            
        print(f"{response.text}:") # Debug
        reason = response.reason
        if result:
            data = response.json()
            if 'result' in data:
                data = data["result"]
                reason = data
            if 'message' in data:
                data = data["message"]
                reason = data
        else:
            data = None 

        if showstatus:
            print(f"{response.status_code}:", reason if reason != "" else "")
        return data

		
# ==========================================================================================================================


    def Login(self, username, password):
        """Returns a JWT login token."""
        self.username = username
        self.password = password
		
        HEADERS = {
        "content-type": "application/json",
        "accept": "application/json",
        "apikey": self.api_key,
        "jwtusername": username,
        "jwtpassword": password
        }		
		
        response = requests.request("POST", self.url.login_url(), headers=HEADERS)
        if response:
            data = json.loads(response.text)
            self.auth_token = data["token"]
		
            if self.auth_token != "":
                self.auth_time = datetime.datetime.now()
                self.HEADERS = {
                "content-type": "application/json",
                "accept": "application/json, text/plain, */*",
                "Accept-Encoding": "gzip, deflate, br",
                "authentication": "Bearer " + self.auth_token,
                "Access-Control-Request-Headers": "authentication"
                }		
            else:
                self.auth_time = none
                self.HEADERS = {}
                # In case of an error one of the following responses is returned:
                #   {"status":"error","classname":null,"message":"User \"demo1\" not found"}
                #   {"status":"error","classname":null,"message":"Wrong password for user \"demo\". 3 more tries left. "}
        else:
            raise Exception(f"Login error: "+response.reason)

        return response

    def Logout(self):
        self.username = None
        self.password = None
        self.auth_token = None
        self.auth_time = None
        self.HEADERS = {}
		
        return 

# ==========================================================================================================================

    def CheckApiLevelOk(self, printmessage = False):
        """Checks if the Wrappers-API-Level matches the Webserver."""
        apilevel = None
        response = self.GetSystem()
        
        if response:
             data = response.json()
             if 'restapilevel' in data:
                 apilevel = data["restapilevel"]
        else:
             raise Exception(f" {response.status_code} {response.reason}")

        result = (apilevel == XTRESTAPILEVEL)
        if result == False:
            raise Exception(f"API-Level missmatch: Client = {XTRESTAPILEVEL} Server = {apilevel}")
        if printmessage:
            print(f"REST-API-Level: OK") 

        return result
		
# ==========================================================================================================================

    def GetSystem(self):
        """Returns global system configuration."""
        return self.__get_data(self.url.system_url())

    def GetUsers(self, params):
        """Returns all of the users."""
        return self.__get_data(self.url.users_url(), params=params)

    def GetUser(self, user_nr):
        """Returns a specific user provided the user_nr (GUID)
            @param self - the object pointer
            @param user_nr - GUID of user
        """
        return self.__get_data(self.url.user_url().format(id=user_nr))

# ==========================================================================================================================

    def GetClients(self, params):
        """Returns all of the clients."""
        return self.__get_data(self.url.clients_url(), params)

    def GetClient(self, client_nr):
        """Returns a specific user provided the client_nr (GUID)
            @param self - the object pointer
            @param client_nr - GUID of client
        """
        return self.__get_data(self.url.client_url().format(id=client_nr))

    def CreateClient(self, data):
        """Creates a new client and returns the client record
            @param self - the object pointer
            @param data - Field values
        """
        return self.__submit_post(self.url.clients_url(), data)

    def UpdateClient(self, client_nr, data):
        """Updates a specific client provided the client_nr (GUID)
            @param self - the object pointer
            @param client_nr - GUID of client
            @param data - Field values
        """
        return self.__put_data(self.url.client_url().format(id=client_nr), data)

    def DeleteClient(self, client_nr):
        """Updates a specific client provided the client_nr (GUID)
            @param self - the object pointer
            @param client_nr - GUID of client
        """
        return self.__delete(self.url.client_url().format(id=client_nr))
    
# ==========================================================================================================================

    def GetProjects(self, params):
        """Returns all of the projects."""
        return self.__get_data(self.url.projects_url(), params)

    def GetProject(self, project_nr):
        """Returns a specific user provided the project_nr (GUID)
            @param self - the object pointer
            @param project_nr - GUID of project
        """
        return self.__get_data(self.url.project_url().format(id=project_nr))

    def CreateProject(self, data):
        """Creates a new project and returns the project record
            @param self - the object pointer
            @param data - Field values
        """
        return self.__submit_post(self.url.projects_url(), data)

    def UpdateProject(self, project_nr, data):
        """Updates a specific project provided the project_nr (GUID)
            @param self - the object pointer
            @param project_nr - GUID of project
            @param data - Field values
        """
        return self.__put_data(self.url.project_url().format(id=project_nr), data)

    def DeleteProject(self, project_nr):
        """Updates a specific project provided the project_nr (GUID)
            @param self - the object pointer
            @param project_nr - GUID of project
        """
        return self.__delete(self.url.project_url().format(id=project_nr))

    def ProjectAssignUser(self, project_nr, user_nr ):
        """Assigns a user to a project
            @param self - the object pointer
            @param project_nr - GUID of project
            @param user_nr - GUID of user
        """
		
        return self.__call_jrpc("ProjectAssignUser", [project_nr, user_nr])
    
    def ProjectUnassignUser(self, project_nr, user_nr ):
        """Unassigns a user from a project
            @param self - the object pointer
            @param project_nr - GUID of project
            @param user_nr - GUID of user
        """
		
        return self.__call_jrpc("ProjectUnassignUser", [project_nr, user_nr])
   

    def ProjectSetProjectstate(self, project_nr, projectstate ):
        """Changes the user of a task
            @param self - the object pointer
            @param project_nr - GUID of project
            @param projectstate - INTEGER projectstate
        """

        return self.__call_jrpc("ProjectSetProjectstate", [project_nr, projectstate ])

    def ProjectGetMinutesNeeded(self, project_nr ):
        """Get the time needed in minutes of a project
            @param self - the object pointer
            @param project_nr - GUID of project
        """

        return self.__call_jrpc("ProjectGetMinutesNeeded", [project_nr ])
    
# ==========================================================================================================================

    def GetTasks(self, params):
        """Returns all of the tasks."""
        return self.__get_data(self.url.tasks_url(), params)

    def GetTask(self, todo_nr):
        """Returns a specific task provided the todo_nr (GUID)
            @param self - the object pointer
            @param todo_nr - GUID of task
        """
        return self.__get_data(self.url.task_url().format(id=todo_nr))

    def CreateTask(self, data):
        """Creates a new task and returns the task record
            @param self - the object pointer
            @param data - Field values
        """
        return self.__submit_post(self.url.tasks_url(), data)

    def UpdateTask(self, todo_nr, data):
        """Updates a specific task provided the todo_nr (GUID)
            @param self - the object pointer
            @param todo_nr - GUID of task
            @param data - Field values
        """
        return self.__put_data(self.url.task_url().format(id=todo_nr), data)

    def DeleteTask(self, todo_nr):
        """Updates a specific task provided the todo_nr (GUID)
            @param self - the object pointer
            @param todo_nr - GUID of task
        """
        return self.__delete(self.url.task_url().format(id=todo_nr))

    def TaskSetUser(self, todo_nr, user_nr ):
        """Changes the user of a task
            @param self - the object pointer
            @param todo_nr - GUID of task
            @param user_nr - GUID of user
        """

        return self.__call_jrpc("TaskSetUser", [todo_nr, user_nr ])

    def TaskSetTaskstate(self, todo_nr, taskstate ):
        """Changes the taskstate of a task
            @param self - the object pointer
            @param todo_nr - GUID of task
            @param taskstate - INTEGER taskstate
        """

        return self.__call_jrpc("TaskSetTaskstate", [todo_nr, taskstate ])

    def TaskGetMinutesNeeded(self, todo_nr ):
        """Get the time needed in minutes of a task
            @param self - the object pointer
            @param todo_nr - GUID of task
        """

        return self.__call_jrpc("TaskGetMinutesNeeded", [todo_nr ])

# ==========================================================================================================================

    def GetTimestamps(self, params):
        """Returns all of the times."""
        return self.__get_data(self.url.timestamps_url(), params)

    def GetTimestamp(self, times_nr):
        """Returns a specific timestamp provided the times_nr (GUID)
            @param self - the object pointer
            @param times_nr - GUID of timestamp
        """
        return self.__get_data(self.url.timestamp_url().format(id=times_nr))

    def CreateTimestamp(self, data):
        """Creates a new timestamp and returns the timestamp record
            @param self - the object pointer
            @param data - Field values
        """
        return self.__submit_post(self.url.timestamps_url(), data)

    def UpdateTimestamp(self, times_nr, data):
        """Updates a specific timestamp provided the times_nr (GUID)
            @param self - the object pointer
            @param times_nr - GUID of timestamp
            @param data - Field values
        """
        return self.__put_data(self.url.timestamp_url().format(id=times_nr), data)

    def DeleteTimestamp(self, times_nr):
        """Updates a specific timestamp provided the times_nr (GUID)
            @param self - the object pointer
            @param times_nr - GUID of timestamp
        """
        return self.__delete(self.url.timestamp_url().format(id=times_nr))

# ==========================================================================================================================

    def GetUserByPersnr(self, persnr):
        """Returns a user_nr of the user with this personal number 
            @param self - the object pointer
            @param persnr - INTEGER personel number
        """
		
        return self.__call_jrpc("GetUserByPersnr", [persnr])

    def GetUserByLoginname(self, loginname):
        """Returns a user_nr of the user with this loginname 
            @param self - the object pointer
            @param loginname - STRING loginname of user
        """
		
        return self.__call_jrpc("GetUserByLoginname", [loginname])

    def GetClientByClientID(self, clientid):
        """Returns a client_nr of the client with this number 
            @param self - the object pointer
            @param clientid - INTEGER client id
        """
		
        return self.__call_jrpc("GetClientByClientID", [clientid])
		
    def GetClientByClientname(self, clientname):
        """Returns a client_nr of the client with this name 
            @param self - the object pointer
            @param clientname - STRING client name
        """
		
        return self.__call_jrpc("GetClientByClientname", [clientname])

    def GetProjectByProjectnumber(self, projectnumber):
        """Returns a project_nr of the project with this project number 
            @param self - the object pointer
            @param projectnumber - STRING projectnumber
        """
		
        return self.__call_jrpc("GetProjectByProjectnumber", [projectnumber])

    def GetTaskByTasknumber(self, tasknumber):
        """Returns a todo_nr of the task with this number 
            @param self - the object pointer
            @param tasknumber - INTEGER tasknumber
        """
		
        return self.__call_jrpc("GetTaskByTasknumber", [tasknumber])
		
# ==========================================================================================================================

    def GetClientByUserdefinedField(self, fieldnumber, search):
        """Returns a client_nr 
            @param self - the object pointer
            @param fieldnumber - INTEGER number of the userdefined field to search into
            @param search - STRING value to search for
        """
		
        return self.__call_jrpc("GetClientByUserdefinedField", [fieldnumber, search])

    def GetProjectByUserdefinedField(self, fieldnumber, search):
        """Returns a project_nr 
            @param self - the object pointer
            @param fieldnumber - INTEGER number of the userdefined field to search into
            @param search - STRING value to search for
        """
		
        return self.__call_jrpc("GetProjectByUserdefinedField", [fieldnumber, search])

    def GetTaskByUserdefinedField(self, fieldnumber, search):
        """Returns a todo_nr 
            @param self - the object pointer
            @param fieldnumber - INTEGER number of the userdefined field to search into
            @param search - STRING value to search for
        """
		
        return self.__call_jrpc("GetTaskByUserdefinedField", [fieldnumber, search])

# ==========================================================================================================================

    def GetClientname(self, client_nr ):
        """Returns the name of the client
            @param self - the object pointer
            @param client_nr - GUID of client
        """
		
        return self.__call_jrpc("GetClientname", [client_nr])

    def GetProjectname(self, project_nr ):
        """Returns the name of the project
            @param self - the object pointer
            @param project_nr - GUID of project
        """
		
        return self.__call_jrpc("GetProjectname", [project_nr])

    def GetTaskname(self, todo_nr ):
        """Returns the name of the task
            @param self - the object pointer
            @param todo_nr - GUID of task
        """
		
        return self.__call_jrpc("GetTaskname", [todo_nr])

    def GetUsername(self, user_nr ):
        """Returns the name of the user
            @param self - the object pointer
            @param user_nr - GUID of user
        """
		
        return self.__call_jrpc("GetUsername", [user_nr])

# ==========================================================================================================================

    def ProjectStart(self, project_nr, user_nr, comment = "" ):
        """Starts a project for a user with optional comment
            @param self - the object pointer
            @param project_nr - GUID of project
            @param user_nr - GUID of user
            @param comment - STRING comment
        """

        starttime = datetime.datetime.now().isoformat()		
        return self.__call_jrpc("TimestampStart", [starttime, "", user_nr, project_nr, "", comment, "", "", "", 0, 0])

    def TaskStart(self, todo_nr, user_nr, comment = "" ):
        """Starts a task for a user with optional comment
            @param self - the object pointer
            @param todo_nr - GUID of task
            @param user_nr - GUID of user
            @param comment - STRING comment
        """

        starttime = datetime.datetime.now().isoformat()		
        return self.__call_jrpc("TimestampStart", [starttime, "", user_nr, "", todo_nr, comment, "", "", "", 0, 0])

    def TimestampStop(self, user_nr, comment = "" ):
        """Stops the running timestamp of a user with optional comment
            @param self - the object pointer
            @param user_nr - GUID of user
            @param comment - STRING comment
        """

        stoptime = datetime.datetime.now().isoformat()		
        return self.__call_jrpc("TimestampStop", [stoptime, user_nr, comment])

